#!/usr/bin/python3

import sys, os
import csv
from pathlib import Path
import datetime

import xml.etree.ElementTree as ET

# Parametry:
#   - cesta k adresáři s rozbalenými XML soubory
#   - cesta k výstupnímu adresáři, kam bude uloženo gtfs

SR70 = {}
for row in csv.DictReader(open('sr70.csv')):
    kod = int(row['SR70'][:-1]) # odebereme koncovou kontrolní číslici
    SR70[kod] = row

dir = Path(sys.argv[1])
out_dir = Path(sys.argv[2])

ids = set()

EXC_ADD = 1
EXC_REMOVE = 2
class Calendar:
    def __init__(self, cal_elem):
        # Načte kalendář a vrátí ho jako množinu objektů typu date
        bitmap = cal_elem.find("BitmapDays").text.strip()
        start = cal_elem.find("ValidityPeriod/StartDateTime").text.strip()
        if not start.endswith('T00:00:00'): raise Exception
        start = datetime.date.fromisoformat(start.split('T')[0])
        if bitmap != '1': # u vlaků, které jedou jen jeden den, chybí EndDateTime
            end = cal_elem.find("ValidityPeriod/EndDateTime").text.strip()
            if not end.endswith('T00:00:00'): raise Exception
            end = datetime.date.fromisoformat(end.split('T')[0])
            if start + datetime.timedelta(days=len(bitmap) - 1) != end: raise Exception("Nesedí EndDateTime")
        r = set()
        for offset, active in enumerate(bitmap):
            if active == '1':
                r.add(start + datetime.timedelta(days=offset))
        self.start = start
        self.end = start + datetime.timedelta(days=offset)
        self.dates = frozenset(r)
        self.bitmap = bitmap

    @property
    def service_interval(self):
        cur = self.start
        while cur <= self.end:
            yield cur
            cur += datetime.timedelta(1)

    def guess_weekdays(self):
        wd_active = {}
        wd_inactive = {}
        for date in self.service_interval:
            active = date in self.dates
            wd = date.weekday()
            dct = wd_active if active else wd_inactive
            dct.setdefault(wd, 0)
            dct[wd] += 1
        return { wd for wd in range(7) if wd_active.get(wd, 0) > wd_inactive.get(wd, 0) }

    def exceptions(self, regular_wd=None):
        if regular_wd is None: regular_wd = self.guess_weekdays()
        r = []
        for date in self.service_interval:
            active = date in self.dates
            wd = date.weekday()
            regular_active = wd in regular_wd
            if active != regular_active:
                r.append((date, EXC_ADD if active else EXC_REMOVE))
        return r


        




calendars = {}
def load_calendar(cal_elem):
    cal = Calendar(cal_elem)
    if cal.dates in calendars: # stejný kalendář (se stejnou množinou dnů) už jsme viděli, zrecyklujeme ho
        return calendars[cal.dates]
    else:
        calendars[cal.dates] = cal
        return cal

ACT_STOP = '0001'
ACT_ON_ONLY = '0028'
ACT_OFF_ONLY = '0029'

KOMERCNI_DRUHY = {
        elem.attrib['KodTAF']: elem.attrib['Kod']
        for elem in ET.parse("komercni_druhy.xml").getroot().findall('.//{http://provoz.szdc.cz/kadr}KomercniDruhVlaku')
}

def parse_timing(elem):
    if elem is None: return None
    val = elem.find('Time').text
    if not val.endswith('.0000000+01:00'): raise Exception
    return datetime.time.fromisoformat(val.split('.')[0])
    

class Train:
    def __init__(self, file):
        tree = ET.parse(file)
        root = tree.getroot()
        pa_elem = root.find("Identifiers/PlannedTransportIdentifiers[ObjectType='PA']")
        tr_elem = root.find("Identifiers/PlannedTransportIdentifiers[ObjectType='TR']")
        pa_core = pa_elem.find("Core").text
        tr_core = tr_elem.find("Core").text
        pa_variant = pa_elem.find("Variant").text
        tr_variant = tr_elem.find("Variant").text
        if pa_core != tr_core or pa_variant != tr_variant:
            raise Exception("PA_ID != TR_ID není podporováno (typicky se vyskytuje u výlukových jízdních řádů) - %s %s %s" % (file, pa_id, tr_id))
        cal_elem = root.find("CZPTTInformation/PlannedCalendar")
        self.calendar = load_calendar(cal_elem)
        self.id = tr_core.strip('-').lstrip('0').rstrip('A') + ('-' + tr_variant.lstrip('0') if int(tr_variant) else '')
        self.id_core = tr_core
        self.id_variant = tr_variant
        stops = []

        number = None
        com_type = None


        for loc in root.findall('CZPTTInformation/CZPTTLocation'):
            code = int(loc.find('LocationPrimaryCode').text)
            country = loc.find('CountryCodeISO').text.upper()
            if country != 'CZ':
                # Pro jednoduchost přeskočíme všechny body mimo území ČR
                continue
            sr70 = SR70[code]
            name = sr70['Tarifní název']
            activities = [ x.text for x in loc.findall('TrainActivity/TrainActivityType') ]
            if ACT_STOP not in activities: continue
            arr = parse_timing(loc.find("TimingAtLocation/Timing[@TimingQualifierCode='ALA']"))
            dep = parse_timing(loc.find("TimingAtLocation/Timing[@TimingQualifierCode='ALD']"))
            if arr is None and dep is not None: arr = dep
            if dep is None and arr is not None: dep = arr
            if number is None: # bereme první číslo vlaku, změny po cestě neřešíme
                number = int(loc.find('OperationalTrainNumber').text)
                com_type = KOMERCNI_DRUHY[loc.find('CommercialTrafficType').text]
            stops.append((code, name,  arr, dep))

        name_elem = root.find("NetworkSpecificParameter[Name='CZTrainName']")
        if name_elem: name = name_elem.find("Value").text
        else: name = ''

        self.number = number
        self.com_type = com_type
        self.name = name
        self.short_name = '%s %s' % (com_type, number)
        if name:
            self.long_name = '%s (%s - %s)' % (name if name else '', stops[0][1], stops[-1][1])
        else:
            self.long_name = '%s - %s' % (stops[0][1], stops[-1][1])

        self.stops = stops


cals_for_core = {}
trains = {}

for file in dir.iterdir():
    train = Train(file)
    if len(train.stops) <= 1: continue # Vlak s jednou zastávkou nemá smysl. Typicky mezinárodní vlak, který stojí na jediném místě v ČR.
    cfc = cals_for_core.setdefault(train.id_core, set())
    if train.calendar.dates & cfc:
        print("VAROVÁNÍ: Překrývající se kalendáře pro varianty core id %s (při přidávání varianty %s), průnik %r" % (train.id_core, train.id_variant, (train.calendar.dates&cfc)))
        continue
    cfc |= train.calendar.dates
    trains[train.id] = train

all_stops = { stop[0] for train in trains.values() for stop in train.stops }

def normalize_name(name):
    return name.replace('hlavní nádraží', 'hl.n.')

def convert_gps(s):
    # převede gps z formátu stupně-minuty-vteřiny na desetinné stupně, které očekává GTFS
    assert s[0] in 'NE'
    deg, rest = s[1:].split('°')
    minutes, seconds = rest.rstrip('"').split("'")
    deg = int(deg)
    minutes = int(minutes or '0')
    seconds = float(seconds.replace(',', '.') or '0')
    return deg + minutes/60 + seconds/3600

# manuální souřadnice pro stanice, kterým v SR70 chybí
GPS_OVERRIDE = {
        # chybí v SR70
        54225: (50.3214320, 14.9444698), # Luštěnice, zdroj: OSM
        54688: (50.2574, 14.5076), # Neratovice sídliště, zdroj: https://cs.wikipedia.org/wiki/Neratovice_s%C3%ADdli%C5%A1t%C4%9B
        # chybné souřadnice v SR70
        54564: (50.1529201, 15.0288512), # Hořátev, zdroj: OSM
        74855: (49.7860106, 13.1299738), # Pňovany zastávka, zdroj: OSM
}

with open(out_dir / 'stops.txt', 'w') as fh:
    wr = csv.DictWriter(fh, ['stop_id', 'stop_name', 'stop_lat', 'stop_lon', 'location_type'])
    wr.writeheader()
    for code in sorted(all_stops):
        sr70 = SR70[code]
        if code in GPS_OVERRIDE: pos = GPS_OVERRIDE[code]
        else: pos = [ str(convert_gps(sr70['GPS %s'%axis])) for axis in 'YX' ]
        wr.writerow({'stop_id': str(code), 'stop_name': normalize_name(sr70['Tarifní název']),
                     'stop_lat': pos[0],
                     'stop_lon': pos[1],
                     'location_type': '0',
                    })

train_list = sorted(trains.values(), key=lambda train: (train.number, train.id_variant))
# Protože vlaky nemají linky v konvenčním smyslu, uděláme pro každý vlak vlastní route
with open(out_dir / 'routes.txt', 'w') as fh:
    wr = csv.DictWriter(fh, ['route_id', 'route_short_name', 'route_long_name', 'route_type'])
    wr.writeheader()
    for train in train_list:
        wr.writerow({'route_id': train.id, 'route_short_name': train.short_name, 'route_long_name': train.long_name, 'route_type': '2'})

# Přiřadíme všem kalendářům identifikační čísla
for idx, cal in enumerate(calendars.values()):
    cal.id = idx

with open(out_dir / 'trips.txt', 'w') as fh:
    wr = csv.DictWriter(fh, ['route_id', 'service_id', 'trip_id'])
    wr.writeheader()
    for train in train_list:
        wr.writerow({'route_id': train.id, 'service_id': str(train.calendar.id), 'trip_id': train.id})

with open(out_dir / 'stop_times.txt', 'w') as fh:
    wr = csv.DictWriter(fh, ['trip_id', 'arrival_time', 'departure_time', 'stop_id', 'stop_sequence'])
    wr.writeheader()
    for train in train_list:
        last_time = None
        jumped_midnight = False
        def gtfs_time(time):
            global last_time, jumped_midnight
            if last_time is not None and time < last_time and not jumped_midnight:
                print("midnigth jump", train.short_name, last_time, time)
                jumped_midnight = True
            last_time = time
            hour = time.hour
            minute = time.minute
            second = time.second
            if jumped_midnight: hour += 24 # časy po půlnoci je třeba zapsat jako e.g. 24:30:00
            return '%d:%02d:%02d'%(hour,minute,second)
        for idx, (code, name, arr, dep) in enumerate(train.stops):
            arr_time = gtfs_time(arr)
            dep_time = gtfs_time(dep)
            wr.writerow({'trip_id': train.id, 'arrival_time': arr_time, 'departure_time': dep_time, 'stop_id': str(code), 'stop_sequence': idx+1})

def gtfs_date(date): return date.strftime('%Y%m%d')
with open(out_dir / 'calendar.txt', 'w') as fh:
    WDAYS = ['monday', 'tuesday', 'wednesday', 'thursday', 'friday', 'saturday', 'sunday']
    wr = csv.DictWriter(fh, ['service_id'] + WDAYS + ['start_date', 'end_date'])
    wr.writeheader()
    for cal in calendars.values():
        row = {'service_id': str(cal.id), 'start_date': gtfs_date(cal.start), 'end_date': gtfs_date(cal.end)}
        regular_wds = cal.guess_weekdays()
        for wd, colname in enumerate(WDAYS):
            row[colname] = int(bool(wd in regular_wds))
        wr.writerow(row)

with open(out_dir / 'calendar_dates.txt', 'w') as fh:
    wr = csv.DictWriter(fh, ['service_id', 'date', 'exception_type'])
    wr.writeheader()
    for cal in calendars.values():
        for date,exc_type in cal.exceptions():
            wr.writerow({'service_id': cal.id, 'date': gtfs_date(date), 'exception_type': str(exc_type)})
